@@ -12,7 +12,7 @@ use File::LibMagic;
 use Carp qw(confess);
 use Data::Dumper;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 our $CLIENT = "Net-JMAPTalk";
 our $AGENT = "$CLIENT/$VERSION";
@@ -44,7 +44,30 @@ sub uploaduri {
   my $scheme = $Self->{scheme} // 'http';
   my $host = $Self->{host} // 'localhost';
   my $port = $Self->{port} // ($scheme eq 'http' ? 80 : 443);
-  my $url = $Self->{uploadurl} // '/upload';
+  my $url = $Self->{uploadurl} // '/jmap/upload';
+
+  return $url if $url =~ m/^http/;
+
+  return "$scheme://$host:$port$url";
+}
+
+sub downloaduri {
+  my $Self = shift;
+  my ($accountId, $blobId, $name) = @_;
+  die "need account and blob" unless ($accountId and $blobId);
+  $name ||= "download";
+  my $scheme = $Self->{scheme} // 'http';
+  my $host = $Self->{host} // 'localhost';
+  my $port = $Self->{port} // ($scheme eq 'http' ? 80 : 443);
+  my $url = $Self->{downloadurl} // '/jmap/download/{accountId}/{blobId}/{name}';
+
+  my %map = (
+    accountId => $accountId,
+    blobId => $blobId,
+    name => $name,
+  );
+
+  $url =~ s/\{([a-zA-Z0-9_]+)\}/$map{$1}||''/ges;
 
   return $url if $url =~ m/^http/;
 
@@ -159,6 +182,10 @@ sub Upload {
     content => $data,
   });
 
+  if ($ENV{DEBUGJMAP}) {
+    warn "JMAP UPLOAD " . Dumper($Response);
+  }
+
   my $jdata;
   $jdata = eval { decode_json($Response->{content}) } if $Response->{success};
 
@@ -175,6 +202,33 @@ sub Upload {
   return $jdata;
 }
 
+sub Download {
+  my $Self = shift;
+  my $cb;
+  if (ref($_[0]) eq 'CODE') {
+    $cb = shift;
+  }
+  my $uri = $Self->downloaduri(@_);
+
+  my %Headers;
+  if ($Self->{user}) {
+    $Headers{'Authorization'} = $Self->auth_header();
+  }
+  if ($Self->{token}) {
+    $Headers{'Authorization'} = "JMAP $Self->{token}";
+  }
+
+  my %getopts = (headers => \%Headers);
+  $getopts{data_callback} = $cb if $cb;
+  my $Response = $Self->ua->get($uri, \%getopts);
+
+  if ($ENV{DEBUGJMAP}) {
+    warn "JMAP DOWNLOAD @_ " . Dumper($Response);
+  }
+
+  die "Failed to download $uri" unless $Response->{success};
+  return $Response->{content};
+}
 
 1;
 __END__
